package com.github.itzilly.sbt.command

object ModCommandRegistry {
    val subcommands = mutableMapOf<String, CommandEntry>()

    fun register(target: Any) {
        for (method in target::class.java.declaredMethods) {
            val annotation = method.getAnnotation(SubCommand::class.java) ?: continue
            val entry = CommandEntry(annotation, method, target)
            subcommands[annotation.name.lowercase()] = entry
            annotation.aliases.forEach {
                subcommands[it.lowercase()] = entry
            }
        }
    }

    fun dispatch(context: CommandContext) {
        val sub = context.arg(0)?.lowercase()
        val subArgs = context.args.drop(1).toTypedArray()

        if (sub == null || sub == "help") {
            showHelp(context, subArgs.firstOrNull())
            return
        }

        val cmd = subcommands[sub]
        if (cmd != null) {
            context.args = subArgs
            cmd.invoke(context)
        } else {
            context.send("Unknown subcommand '$sub'. Use /sbt help.")
        }
    }

    private fun showHelp(ctx: CommandContext, target: String?) {
        if (target != null) {
            val cmd = subcommands[target.lowercase()]
            if (cmd != null) {
                ctx.send("Usage: /sbt ${cmd.annotation.name} ${cmd.annotation.usage}")
                ctx.send(cmd.annotation.description)
                return
            }
            ctx.send("Unknown command '$target'")
            return
        }

        ctx.send("Available /sbt commands:")
        subcommands.values.toSet().forEach {
            ctx.send("- ${it.annotation.name} ${it.annotation.usage} : ${it.annotation.description}")
        }
    }

    class CommandEntry(
        val annotation: SubCommand,
        val method: java.lang.reflect.Method,
        val target: Any
    ) {
        fun invoke(ctx: CommandContext) {
            try {
                method.invoke(target, ctx)
            } catch (e: Exception) {
                ctx.send("§cError: ${e.cause?.message ?: e.message}")
            }
        }
    }
}