package com.github.itzilly.sbt.features

import com.github.itzilly.sbt.Keybinds
import com.github.itzilly.sbt.render.BlockOutlineRenderer
import com.github.itzilly.sbt.util.Chatterbox
import com.github.itzilly.sbt.util.DelayedFunction
import com.github.itzilly.sbt.util.SimpleChatMsg
import com.github.itzilly.sbt.util.getMiniServerIdFromScoreboard
import net.minecraft.block.Block
import net.minecraft.client.Minecraft
import net.minecraft.init.Blocks
import net.minecraft.util.BlockPos
import net.minecraft.util.ChatComponentText
import net.minecraft.util.ChatStyle
import net.minecraft.util.EnumChatFormatting
import net.minecraft.world.World
import net.minecraft.world.chunk.Chunk
import net.minecraftforge.client.event.RenderWorldLastEvent
import net.minecraftforge.event.entity.EntityJoinWorldEvent
import net.minecraftforge.event.world.ChunkEvent
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent
import net.minecraftforge.fml.common.gameevent.InputEvent
import net.minecraftforge.fml.common.gameevent.TickEvent

object GrottoFinder {
    private var isRunningAutoDetect: Boolean = true
    private var autoDetectEnabled: Boolean = false
    private val scannedChunks = mutableMapOf<Pair<Int, Int>, Long>()
    private val pendingJasperBlocks = mutableListOf<BlockPos>()
    private var hasLocatedGrotto = false
    private const val SCAN_COOLDOWN_MS = 5 * 60 * 1100 // 5 minutes (+ a little to account for lag)

    // World Bounds 824, 189, 824 201, 30, 201
    private val minBounds = BlockPos(201, 30, 201)  // Max area of the Crystal Hollows
    private val maxBounds = BlockPos(824, 189, 824) // Max area of the Crystal Hollows
    private val targetBlock: Block = Blocks.stained_glass
    private val targetMeta: Int = 2 // Magenta stained glass

    private var lastWorldLoadTime = 0L
    private var lastMiniServerId: String? = null

    private var renderBlockList: ArrayList<BlockPos> = ArrayList()
    private var jasperCrystalsList: ArrayList<BlockPos> = ArrayList()

    fun enableAutoDetect() {
        autoDetectEnabled = true
    }

    fun disableAutoDetect() {
        autoDetectEnabled = false
    }

    fun isAutoDetectEnabled(): Boolean {
        return autoDetectEnabled
    }

    private fun clear() {
        renderBlockList.clear()
    }

    private fun magentaNodeExists(x: Int, y: Int, z: Int): Boolean {
        return renderBlockList.any { it.x == x && it.y == y && it.z == z }
    }

    @SubscribeEvent
    fun onKeyInput(event: InputEvent.KeyInputEvent) {
        if (Keybinds.searchGrotto.isPressed) {
            Chatterbox.say("Checking chunks for grotto...")
            checkWorldForMagentaStainedGlassAsync(Minecraft.getMinecraft().theWorld)
            renderBlockList.addAll(jasperCrystalsList)
        }

        if (Keybinds.clearGrotto.isPressed) {
            renderBlockList.clear()
        }
    }

    @Deprecated("Use async instead")
    private fun checkWorldForMagentaStainedGlassSync(world: World) {
        for (chunkX in minBounds.x / 16..maxBounds.x / 16) {
            for (chunkZ in minBounds.z / 16..maxBounds.z / 16) {
                val chunk = world.getChunkFromChunkCoords(chunkX, chunkZ)

                // Iterate through the chunk's blocks within the world bounds
                for (x in 0..15) {
                    for (y in minBounds.y..maxBounds.y) {
                        for (z in 0..15) {
                            val worldX = chunk.xPosition * 16 + x
                            val worldZ = chunk.zPosition * 16 + z

                            // Ensure the block is within the specified bounds
                            if (worldX in minBounds.x..maxBounds.x && worldZ in minBounds.z..maxBounds.z) {
                                val blockPos = BlockPos(worldX, y, worldZ)
                                val blockState = chunk.getBlockState(blockPos)
                                val block = blockState.block
                                val meta = block.getMetaFromState(blockState)

                                if (block === targetBlock && meta == targetMeta) {
                                    if (!magentaNodeExists(blockPos.x, blockPos.y, blockPos.z)) {
                                        renderBlockList.add(BlockPos(blockPos.x, blockPos.y, blockPos.z))
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        if (renderBlockList.isNotEmpty()) {
            val pos = renderBlockList[0]
            val msg = SimpleChatMsg(
                "Found Jasper near (%d, %d, %d)".format(
                    pos.x,
                    pos.y,
                    pos.z
                )
            ).light_purple()
            Chatterbox.say(msg)
        }
    }

    private fun checkWorldForMagentaStainedGlassAsync(world: World) {
        Thread {
            val foundBlocks = ArrayList<BlockPos>()
            for (chunkX in minBounds.x / 16..maxBounds.x / 16) {
                for (chunkZ in minBounds.z / 16..maxBounds.z / 16) {
                    val chunk = world.getChunkFromChunkCoords(chunkX, chunkZ)

                    for (x in 0..15) {
                        for (y in minBounds.y..maxBounds.y) {
                            for (z in 0..15) {
                                val worldX = chunk.xPosition * 16 + x
                                val worldZ = chunk.zPosition * 16 + z

                                if (worldX in minBounds.x..maxBounds.x && worldZ in minBounds.z..maxBounds.z) {
                                    val blockPos = BlockPos(worldX, y, worldZ)
                                    val blockState = chunk.getBlockState(blockPos)
                                    val block = blockState.block
                                    val meta = block.getMetaFromState(blockState)

                                    if (block === targetBlock && meta == targetMeta) {
                                        foundBlocks.add(blockPos)
                                    }
                                }
                            }
                        }
                    }
                }
            }

            Minecraft.getMinecraft().addScheduledTask {
                clear()
                renderBlockList.addAll(foundBlocks)

                if (renderBlockList.isNotEmpty()) {
                    val pos = renderBlockList[0]
                    val msg = SimpleChatMsg(
                        "Found Jasper near (%d, %d, %d)".format(
                            pos.x,
                            pos.y,
                            pos.z
                        )
                    ).light_purple()
                    Chatterbox.say(msg)
                } else {
                    Chatterbox.say("No Jasper found.")
                }
            }
        }.start()
    }

    @SubscribeEvent
    fun renderBlockOverlay(event: RenderWorldLastEvent) {
        val player = Minecraft.getMinecraft().thePlayer

        for (glass: BlockPos in renderBlockList) {
            val pos = BlockPos(glass.x, glass.y, glass.z)
            BlockOutlineRenderer.drawBlockOutline(player, pos, event.partialTicks)
            BlockOutlineRenderer.drawBlockLine(pos, event.partialTicks)
        }
    }

    private fun isInCrystalHollowsFromTab(): Boolean {
        val infoList = Minecraft.getMinecraft().thePlayer?.sendQueue?.playerInfoMap ?: return false

        val lines = infoList
            .mapNotNull { it.displayName?.unformattedText }
            .map { it.trim() }
            .filter { it.isNotBlank() }

        for (i in 0 until lines.size - 1) {
            if (lines[i].equals("Info", ignoreCase = true) &&
                lines[i + 1].contains("Area: Crystal Hollows", ignoreCase = true)
            ) {
                return true
            }
        }

        return false
    }

//    @SubscribeEvent
//    fun onEntityJoinWorldEvent(event: EntityJoinWorldEvent) {
//        val mc = Minecraft.getMinecraft()
//        val player = mc.thePlayer
//        val TWO_SECONDS = 2000
//
//        if (event.entity === player && event.world.isRemote) {
//            val now = System.currentTimeMillis()
//            if (now - lastWorldLoadTime < TWO_SECONDS) return
//            lastWorldLoadTime = now
//
//            DelayedFunction({
//                isRunningAutoDetect = isInCrystalHollowsFromTab()
//                if (isRunningAutoDetect) {
//                    if (pendingJasperBlocks.isNotEmpty()) {
//                        jasperCrystalsList.addAll(pendingJasperBlocks)
//                        pendingJasperBlocks.clear()
//                        showRevealMessage()
//                    }
//                }
//            }, delayTicks = 100)
//        }
//    }

//    @SubscribeEvent
//    fun onClientTick(event: TickEvent.ClientTickEvent) {
//        if (!autoDetectEnabled || event.phase != TickEvent.Phase.END) return
//
//        val currentMiniId = getMiniServerIdFromScoreboard() ?: return
//        if (currentMiniId != lastMiniServerId) {
//            if (lastMiniServerId != null) {
//                // TODO: fix
////                isRunningAutoDetect = false
//                renderBlockList.clear()
//                jasperCrystalsList.clear()
//                pendingJasperBlocks.clear()
//                hasLocatedGrotto = false
//            }
//            lastMiniServerId = currentMiniId
//        }
//    }

//    @SubscribeEvent
//    fun onChunkLoad(event: ChunkEvent.Load) {
//        if (!autoDetectEnabled) return
//
//        val chunk = event.chunk
//        val key = Pair(chunk.xPosition, chunk.zPosition)
//        val now = System.currentTimeMillis()
//
//        if (now - (scannedChunks[key] ?: 0) < SCAN_COOLDOWN_MS) return
//        scannedChunks[key] = now
//
//        scanChunkForJasper(chunk)
//    }

    private fun scanChunkForJasper(chunk: Chunk) {
        Thread {
            val found = mutableListOf<BlockPos>()

            val baseX = chunk.xPosition * 16
            val baseZ = chunk.zPosition * 16

            for (x in 0..15) {
                for (y in minBounds.y..maxBounds.y) {
                    for (z in 0..15) {
                        val pos = BlockPos(baseX + x, y, baseZ + z)
                        val blockState = chunk.getBlockState(pos)
                        val block = blockState.block
                        val meta = block.getMetaFromState(blockState)

                        if (block == targetBlock && meta == targetMeta) {
                            found.add(pos)
                        }
                    }
                }
            }

            if (found.count() != 1) {
                if (isRunningAutoDetect) {
                    jasperCrystalsList.addAll(found)
                    if (!hasLocatedGrotto) {
                        showRevealMessage()
                        hasLocatedGrotto = true
                    }
                } else {
                    pendingJasperBlocks.addAll(found)
                }

            }

            Chatterbox.say("Scanned Chunk, found: ${found.count()}")
        }.start()
    }

    private fun showRevealMessage() {
        Chatterbox.say(SimpleChatMsg("- - - - - - - - - - - - - - - - - - -").light_purple())
        val clickMsg = ChatComponentText("§dDetected Fairy Grotto! §7[Click to reveal]").apply {
            chatStyle = ChatStyle().apply {
                color = EnumChatFormatting.LIGHT_PURPLE
                underlined = true
                chatClickEvent = net.minecraft.event.ClickEvent(
                    net.minecraft.event.ClickEvent.Action.RUN_COMMAND,
                    "/sbt reveal_grotto"
                )
                chatHoverEvent = net.minecraft.event.HoverEvent(
                    net.minecraft.event.HoverEvent.Action.SHOW_TEXT,
                    ChatComponentText("§7Click to show block outlines.")
                )
            }
        }
        Chatterbox.say(clickMsg)
        Chatterbox.say(SimpleChatMsg("- - - - - - - - - - - - - - - - - - -").light_purple())
    }

    fun revealDetectedJasperBlocks() {
        jasperCrystalsList.addAll(pendingJasperBlocks)
        if (jasperCrystalsList.isEmpty()) {
            Chatterbox.say(SimpleChatMsg("[Grotto] No Jasper found!").light_purple())
            return
        }

        renderBlockList.clear()
        renderBlockList.addAll(jasperCrystalsList)
        Chatterbox.say(SimpleChatMsg("[Grotto] Revealed detected Fairy Grotto blocks").light_purple())

        val pos = renderBlockList[0]
        val msg = SimpleChatMsg(
            "Found Jasper near (%d, %d, %d)".format(
                pos.x,
                pos.y,
                pos.z
            )
        ).light_purple()
        Chatterbox.say(msg)
    }
}
